#include <math.h>
#include <stdlib.h>

#if defined(__APPLE__)
  #include <OpenGL/gl.h>
  #include <OpenGL/glu.h>
  #include <GLUT/glut.h>
#else
  #if defined(WIN32) || defined(_WIN32) || defined(__WIN32__)
    #include <windows.h>
  #endif
  #include <GL/gl.h>
  #include <GL/glu.h>
  #include <GL/glut.h>
#endif

#ifndef M_PI
  #define M_PI 3.14159265359
#endif

struct Vector {
  float x, y, z;
 
  Vector(float v = 0) : x(v), y(v), z(v) { }
  Vector(float x, float y, float z) : x(x), y(y), z(z) { }
  Vector operator+(const Vector& v) const { return Vector(x + v.x, y + v.y, z + v.z); }
  Vector operator-(const Vector& v) const { return Vector(x - v.x, y - v.y, z - v.z); }
  Vector operator*(const Vector& v) const { return Vector(x * v.x, y * v.y, z * v.z); }
  Vector operator/(const Vector& v) const { return Vector(x / v.x, y / v.y, z / v.z); }
  friend Vector operator+(float f, const Vector& v) { return v+f; }
  friend Vector operator-(float f, const Vector& v) { return Vector(f)-v; }
  friend Vector operator*(float f, const Vector& v) { return v*f; }
  friend Vector operator/(float f, const Vector& v) { return Vector(f)/v; }
  Vector& operator+=(const Vector& v) { x += v.x, y += v.y, z += v.z; return *this; }
  Vector& operator-=(const Vector& v) { x -= v.x, y -= v.y, z -= v.z; return *this; }
  Vector& operator*=(const Vector& v) { x *= v.x, y *= v.y, z *= v.z; return *this; }
  Vector& operator/=(const Vector& v) { x /= v.x, y /= v.y, z /= v.z; return *this; }
  Vector operator-() const { return Vector(-x, -y, -z); }
  float dot(const Vector& v) const { return x*v.x + y*v.y + z*v.z; }
  friend float dot(const Vector& a, const Vector& b) { return a.dot(b); }
  Vector cross(const Vector& v) const { return Vector(y*v.z - z*v.y, z*v.x - x*v.z, x*v.y - y*v.x); }
  friend Vector cross(const Vector& a, const Vector& b) { return a.cross(b); }
  float length() const { return sqrt(x*x + y*y + z*z); }
  Vector normalize() const { float l = length(); if(l > 1e-3) { return (*this/l); } else { return Vector(); } }
  bool isNull() const { return length() < 1e-3; }
};

enum ControllKeys {W, A, S, D, keys_num};
bool keys_down[keys_num];

struct Camera {
  Vector fwd, pos;
  const float speed, mouse_speed;

  Camera(float speed = 5, float mouse_speed = 0.002f) : fwd(Vector(9, 0, 4).normalize()), pos(-9, 5, -4), speed(speed), mouse_speed(mouse_speed) { }

  void updatePos(float dt) {
    Vector up = Vector(0, 1, 0), right = cross(fwd, up).normalize();
    up = cross(right, fwd).normalize();

    if(keys_down[W] && !keys_down[S]) {
      pos += fwd * speed * dt;
    } else if(keys_down[S] && !keys_down[W]) {
      pos -= fwd * speed * dt;
    }

    if(keys_down[D] && !keys_down[A]) {
      pos += right * speed * dt;
    } else if(keys_down[A] && !keys_down[D]) {
      pos -= right * speed * dt;
    }
  }

  void updateDir(int dx, int dy) {
    Vector y_axis = Vector(0, 1, 0), right = cross(fwd, y_axis).normalize();
    Vector up = cross(right, fwd).normalize();

    // Ha teljesen felfele / lefele néznénk, akkor ne forduljon át a kamera
    float dot_up_fwd = dot(y_axis, fwd);
    if(dot_up_fwd > 0.95f && dy > 0) {
      dy = 0;
    } 
    if(dot_up_fwd < -0.95f && dy < 0) {
      dy = 0;
    }

    // Módosítsuk az nézeti irányt
    fwd += mouse_speed * (right * dx + up * dy);
    fwd = fwd.normalize();
  }

  void applyMatrix() const {
    gluLookAt(pos.x, pos.y, pos.z, pos.x+fwd.x, pos.y+fwd.y, pos.z+fwd.z, 0, 1, 0);
  }
} camera;

void glVertex3f(const Vector& v) {
  glVertex3f(v.x, v.y, v.z);
}

void glQuad(const Vector& a, const Vector& b, const Vector& c, const Vector& d) {
  Vector normal = cross(b-a, c-a).normalize();
  glNormal3f(normal.x, normal.y, normal.z);
 
  int res = 10; // resolution
  for(int i = 0; i < res; i++) {
    for(int j = 0; j < res; j++) {
      glVertex3f(a + i*(b-a)/res     + j*(d-a)/res); 
      glVertex3f(a + (i+1)*(b-a)/res + j*(d-a)/res); 
      glVertex3f(a + (i+1)*(b-a)/res + (j+1)*(d-a)/res); 
      glVertex3f(a + i*(b-a)/res     + (j+1)*(d-a)/res); 
    }
  }
}

void drawCube(const Vector& size) {
  glBegin(GL_QUADS); {
    /*       (E)-----(A)
             /|      /|
            / |     / |
          (F)-----(B) |
           | (H)---|-(D)
           | /     | /
           |/      |/
          (G)-----(C)        */

    Vector s = size / 2;

    Vector A(+s.x, +s.y, -s.z), B(+s.x, +s.y, +s.z), C(+s.x, -s.y, +s.z), D(+s.x, -s.y, -s.z), 
           E(-s.x, +s.y, -s.z), F(-s.x, +s.y, +s.z), G(-s.x, -s.y, +s.z), H(-s.x, -s.y, -s.z);

    glQuad(A, B, C, D); glQuad(E, H, G, F); glQuad(A, E, F, B);
    glQuad(D, C, G, H); glQuad(B, F, G, C); glQuad(A, D, H, E);

  } glEnd();
}

int anim_time = 0;

GLfloat black[4] = {0, 0, 0, 1};

void setSun() {
  float p[4] = {-1, 0.8, -0.6, 0};
  glLightfv(GL_LIGHT7, GL_POSITION, p);

  float c[4] = {0.5, 0.5, 0.5, 1};
  glLightfv(GL_LIGHT7, GL_DIFFUSE, c);
  glLightfv(GL_LIGHT7, GL_AMBIENT, black);

  glEnable(GL_LIGHT7);
}


void setLighting(int i) {
  float p[4] = {0.0f, 0.0f, 0.0f, 1};
  glLightfv(GL_LIGHT0 + i, GL_POSITION, p);
  glLightf(GL_LIGHT0 + i, GL_CONSTANT, 0.0f);
  glLightf(GL_LIGHT0 + i, GL_QUADRATIC_ATTENUATION, 1.0f);

  GLfloat c[4] = {0.3f, 0.3f, 0.3f, 1};
  glLightfv(GL_LIGHT0 + i, GL_DIFFUSE, c);
  glLightfv(GL_LIGHT0 + i, GL_AMBIENT, black);
  glEnable(GL_LIGHT0 + i);
}

void drawLamps() {
  float time = 4*anim_time*2*M_PI/200.0f;
  Vector pos[3] = {
    Vector(1.5f*cos(time), 0.5f, 1.5*sin(time)),
    Vector(1.5*cos(1.07f*time + M_PI_2), 1.5f*sin(time + M_PI_2), 0.5f),
    Vector(0.5f, -1.5*sin(1.1f*time), -1.5f*cos(1.05f*time))
  };

  for(int i = 0; i < sizeof(pos)/sizeof(pos[0]); ++i) {
    glPushMatrix(); {
      glTranslatef(pos[i].x, pos[i].y, pos[i].z);
      setLighting(i+1);
      glColor3f(1, 1, 1);
      glutSolidSphere(0.1f, 8, 8);
    } glPopMatrix();
  }
}

GLUquadric *quad = gluNewQuadric();

struct Tree {
  float size;
  Vector pos;
  bool valid;

  void draw() {
    glPushMatrix(); {
      glTranslatef(pos.x, pos.y, pos.z);

      glColor3f(0x53/255., 0x35/255., 0x0A/255.);
      glPushMatrix(); {
        glRotatef(-90, 1, 0, 0);
        gluCylinder(quad, size/3, size/3, size, 8, 8);
      } glPopMatrix();

      glColor3f(0.01, 0.8, 0.12);
      glPushMatrix(); {
        glTranslatef(0, size, 0);
        glRotatef(-90, 1, 0, 0);
        glutSolidCone(size/2, 2*size, 8, 8);
      } glPopMatrix();
    } glPopMatrix();
  }
};

struct Forest {
  Tree trees[13][13];
  Forest() {
    for(int i = 0; i < 13; i++)
      for(int j = 0; j < 13; j++) {
        float x = i*15 - 90, y = j*15 - 90;
        if(fabs(x) < 50 && fabs(y) < 50) {
          trees[i][j].valid = false;
          continue;
        }
        trees[i][j].size = 4 + 3.0f*rand()/RAND_MAX; 
        trees[i][j].pos = Vector(x + 10.0f*rand()/RAND_MAX - 5, 0, y + 10.0f*rand()/RAND_MAX - 5);
        trees[i][j].valid = true;
      }
  }

  void draw() {
    for(int i = 0; i < 13; i++) {
      for(int j = 0; j < 13; j++) {
        if(trees[i][j].valid)
          trees[i][j].draw();
      }
    }
  }
} forest;

void drawGround() {
  glColor3f(0.01, 0.8, 0.12);

  glPushMatrix(); {
    glTranslatef(0, -1.0f, 0);
    glScalef(100, 0.5, 100);
    drawCube(2.0f);
  } glPopMatrix();

  forest.draw();
}

void drawPantheon() {
  drawGround();

  glColor3f(1, 1, 1);

  glPushMatrix(); {
    glTranslatef(0, -1.0f, 0);
    glScalef(30, 1, 30);
    drawCube(2.0f);

    glTranslatef(0, 20, 0);
    drawCube(2.0f);

  } glPopMatrix();

  glColor3f(0.2f, 0.4f, 0.6f);

  glPushMatrix(); {
    glRotatef(-90, 1, 0, 0);
    gluCylinder(quad, 2, 1.2f, 3.8f, 16, 16);
  } glPopMatrix();

  glColor3f(1, 1, 1);

  int num_cylinders = 6;
  for(int i = 0; i < num_cylinders; ++i) {
    for(int j = 0; j < 4; j++) {
      int sign = j%2 ? 1 : -1;

      glPushMatrix(); {
        if(j < 2) {
          glTranslatef(sign * 25, 0, sign * (50.0f*i/num_cylinders - 25));
        } else {
          glTranslatef(-sign * (50.0f*i/num_cylinders - 25), 0, sign * 25);
        }
        
        glRotatef(-90, 1, 0, 0);
        gluCylinder(quad, 2, 2, 20, 8, 8);
      } glPopMatrix();
    }
  }
}

void onDisplay() {
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  camera.applyMatrix();
  setSun();

  glPushMatrix(); {
    glColor3f(1, 1, 0);
    glTranslatef(0, 5.0f, 0);
    glutSolidTeapot(1.0f);
    drawLamps();
  } glPopMatrix();

  drawPantheon();


  ++anim_time;

  glutSwapBuffers();
}

void onIdle() {
  static float last_time = glutGet(GLUT_ELAPSED_TIME);
  float time = glutGet(GLUT_ELAPSED_TIME);
  float dt = (time - last_time) / 1000.0f;
  last_time = time;

  camera.updatePos(dt);
  glutPostRedisplay(); 
}

void onInitialization() {
  glClearColor(135./255., 206./255., 235./255., 1);
  glEnable(GL_DEPTH_TEST);
  glMatrixMode(GL_PROJECTION);
  gluPerspective(60, 1, 0.2, 200);
  glMatrixMode(GL_MODELVIEW);

  glEnable(GL_LIGHTING);
  glEnable(GL_COLOR_MATERIAL);
  glEnable(GL_SMOOTH);
}

void onKeyboard(unsigned char key, int, int) {
  switch(key) {
    case 'w': case 'W':
      keys_down[W] = true;
      break;
    case 's': case 'S':
      keys_down[S] = true;
      break;
    case 'a': case 'A':
      keys_down[A] = true;
      break;
    case 'd': case 'D':
      keys_down[D] = true;
      break;
  }
}

void onKeyboardUp(unsigned char key, int, int) {
    switch(key) {
    case 'w': case 'W':
      keys_down[W] = false;
      break;
    case 's': case 'S':
      keys_down[S] = false;
      break;
    case 'a': case 'A':
      keys_down[A] = false;
      break;
    case 'd': case 'D':
      keys_down[D] = false;
      break;
  }
}

int last_x, last_y;
void onMouse(int, int, int x, int y) {
  last_x = x;
  last_y = y;
}

void onMouseMotion(int x, int y) {
  camera.updateDir(x-last_x, last_y-y);
  last_x = x;
  last_y = y;
}

int main(int argc, char **argv) {
  glutInit(&argc, argv);
  glutInitWindowSize(600, 600);
  glutInitWindowPosition(100, 100);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);

  glutCreateWindow("Grafika pelda program");

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();

  onInitialization();

  glutDisplayFunc(onDisplay);
  glutMouseFunc(onMouse);
  glutIdleFunc(onIdle);
  glutKeyboardFunc(onKeyboard);
  glutKeyboardUpFunc(onKeyboardUp);
  glutMotionFunc(onMouseMotion);

  glutMainLoop();

  return 0;
}
